unit gapfiles;

{*************TMainForm************
Reader and writer class for .GAP files,
Game Addon Packs. Developed for use with Final Fantasy 8.
See http://sfff.cjb.net for more info.

Feel free to use this class if you do something similar with
another game.

The class structure was put toghether in a hurry, if you see
any improvements, please contect me:
dagsverre@sensewave.com

Made in Delphi 4.

Author:  Dag Sverre Seljebotn, dagsverre@sensewave.com
Created: 12-04-2000
History: //20-03-2000 First release

This code is distributed under the GNU General Public License.
NO WARRANTY. See COPYING for details.

Copyright (c) 2000 Dag Sverre Seljebotn
}


interface
  uses SysUtils, Classes, ZLib;

type

{ Layout:
  The whole file is ZLib'ed.

  1 x TGapHeader

  for each file:
  1 x TGapFile with data
  repeat...
}

  TGapFileHeader = record
    HeaderSize,   //This is included so that backwards compatibility will be possible, should not include Data
    FileSize: Cardinal;
    Description: String[255];
    RelativePath: String[255];
    InternalData: Cardinal;
    FileType,
    CRC,
    Flags: Cardinal;   //Currently only CRC in DIFF...
  end;

  PGapFileHeader = ^TGapFileHeader;

  TGapHeader = record
    HeaderSize,              //Again included for possibility of backwards compatibility
    Game: Cardinal;
    Description: String[255];
  end;


  TGapPackageWriter = class
  private
    FFileStream: TFileStream;
    FCompressStream: TCompressionStream;
  protected
    procedure WriteHeader(Header: TGapHeader);
  public
    constructor Create(
      const Filename: string;
      Game: Cardinal;
      Description: String); overload;

    constructor Create(const Filename: string; Header: TGapHeader); overload;
    destructor Destroy; override;

    procedure AddFile(
      Stream         : TStream;
      ByteToWrite    : Cardinal;
      Description,
      RelativePath  : String;
      FileType,
      CRC,
      Flags          : Cardinal); overload;
    procedure AddFile(Header: TGapFileHeader; Stream: TStream); overload;
  end;


  TRegisterPack = procedure(const Game: Cardinal; Description: String) of object;
  TRegisterFile = procedure(const Path, Description: String; FileType, Flags: Cardinal) of object;

  TGapPackageReader = class
  private
    FOnRegisterFile: TRegisterFile;
    FOnRegisterPack: TRegisterPack;
    FFileStream: TFileStream;
    FDeCompressStream: TDeCompressionStream;
  protected

  public
    constructor Create(const Filename: string);
    destructor Destroy; override;

    procedure Extract(const Path: String);
    property OnRegisterPack: TRegisterPack read FOnRegisterPack write FOnRegisterPack;
    property OnRegisterFile: TRegisterFile read FOnRegisterFile write FOnRegisterFile;
  end;

const
  FlagDiff = $1;


implementation

{ TGapPackageWriter }

procedure TGapPackageWriter.AddFile(
  Stream         : TStream;
  ByteToWrite    : Cardinal;
  Description,
  RelativePath  : String;
  FileType,
  CRC,
  Flags          : Cardinal);
var
  GapFile: TGapFileHeader;
begin
  FillChar(GapFile, SizeOf(GapFile), 0);
  GapFile.HeaderSize     := SizeOf(GapFile);
  GapFile.FileSize       := ByteToWrite;
//  GapFile.CompressMethod := Compression;
  GapFile.RelativePath   := RelativePath;
  GapFile.Description    := Description;
  GapFile.FileType       := FileType;
  GapFile.CRC            := CRC;
  GapFile.Flags          := Flags;

  AddFile(GapFile, Stream);
end;

procedure TGapPackageWriter.AddFile(Header: TGapFileHeader;
  Stream: TStream);
begin
  with FCompressStream do
  begin
    WriteBuffer(Header, Header.HeaderSize);
    CopyFrom(Stream, Header.FileSize);
  end;
end;

constructor TGapPackageWriter.Create(const Filename: string;
  Game: Cardinal; Description: String);
var
  Header: TGapHeader;
begin
  FillChar(Header, SizeOf(Header), 0);
  Header.HeaderSize:=SizeOf(Header);
  Header.Game:=Game;
  Header.Description:=Description;
  Create(Filename, Header);
end;

constructor TGapPackageWriter.Create(const Filename: string;
  Header: TGapHeader);
begin
  inherited Create;
  FFileStream:=TFileStream.Create(Filename, fmCreate);
  FCompressStream:=TCompressionStream.Create(clMax, FFileStream);
  WriteHeader(Header);
end;

destructor TGapPackageWriter.Destroy;
begin
  FCompressStream.Free;
  FFileStream.Free;
  inherited Destroy;
end;

procedure TGapPackageWriter.WriteHeader(Header: TGapHeader);
begin
  FCompressStream.Write(Header, Header.HeaderSize);
end;

{ TGapPackageReader }

constructor TGapPackageReader.Create(const Filename: string);
begin
  inherited Create;
  FFileStream:=TFileStream.Create(Filename, fmOpenRead or fmShareDenyNone);
  FDeCompressStream:=TDeCompressionStream.Create(FFileStream);
end;

destructor TGapPackageReader.Destroy;
begin
  FDeCompressStream.Free;
  FFileStream.Free;
  inherited Destroy;
end;

procedure TGapPackageReader.Extract(const Path: String);
var
  MainHeader: TGapHeader;
  FileHeader: TGapFileHeader;
  FilePath: String;
begin
  with FDeCompressStream do
  begin
    ReadBuffer(MainHeader.HeaderSize, 4);
    ReadBuffer(MainHeader.Game, MainHeader.HeaderSize - 4);

    while True do
    begin
      if Read(FileHeader.HeaderSize, 4) < 4 then
        Break;
      if Read(FileHeader.FileSize, FileHeader.HeaderSize - 4) < FileHeader.HeaderSize - 4 then
        Break;
      FilePath:=Path + FileHeader.RelativePath;
      with TFileStream.Create(FilePath, fmCreate) do
      try
        if CopyFrom(FDeCompressStream, FileHeader.FileSize) < FileHeader.FileSize then
          Break;
      finally
        Free;
      end;
      if Assigned(FOnRegisterFile) then
        FOnRegisterFile(FilePath, FileHeader.Description, FileHeader.FileType, FileHeader.Flags);
    end;

    if Assigned(FOnRegisterPack) then
      FOnRegisterPack(MainHeader.Game, MainHeader.Description);
  end;
end;

end.
